/*
 * ata_rtl8012.c
 *
 * Copyright (c)2005 Tobias Hagemann.
 * Author may be reached at <mail@tobiashagemann.de>.
 *
 * This software may be used and distributed according to the terms
 * of the GNU Public License, incorporated herein by reference.
 *
 * This is a network driver for ethernet pocket adapters (pocket
 * means "attached to the cartridge port of an atari") with RTL8012 chip (most
 * adapters have this chip).
 */

static const char *version=
		"ata_rtl8012.c: v0.90 14/06/2005 Tobias Hagemann <mail@tobiashagemann.de>\n";

/*
 * This driver is derived from rtl8012.c written by S O L A M Y L W A R E ,  Stepan Skroband
 * and rtl8012.c from FreeMiNT written by Frank Naumann (http://sparemint.atariforge.net/sparemint/)
 * and Vassilis Papathanassiou (http://users.otenet.gr/~papval/).
 */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
/* We scheduale tasks here because we have no interrupt at the cartridge port*/
#include <linux/tqueue.h>
/* We also need the ability to put ourselves to sleep and wake up later */
#include <linux/sched.h>

#include <asm/system.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

//#define RTL_DEBUG

#define wait5ms		mdelay (5)

/* PAGE 0 of EPLC registers */
#define IDR0		0x0 /*network node ID registers 0-5*/
#define IDR1		0x1
#define IDR2		0x2
#define IDR3		0x3
#define IDR4		0x4
#define IDR5		0x5
#define TBCR0		0x6 /*W-transmit byte count register 0*/
#define TBCR1		0x7 /*W-transmit byte count register 1*/
#define TSR		0x8 /*transmit status register*/
#define RSR		0x9 /*receive status register*/
#define ISR		0xA /*interrupt status register*/
#define IMR		0xB /*interrupt mask register*/
#define CMR1		0xC /*command register 1*/
#define CMR2		0xD /*command register 2*/
#define MODSEL		0xE /*mode select register*/
#define MODDEL		0xF /*mode delay register*/

#define PNR		TBCR0 /*R-received packet number register*/
#define COLR		TBCR1 /*R-collision number register*/

/* PAGE 1 of EPLC registers */
#define MAR0		0x0 /*multicast address registers 0-7*/
#define MAR1		0x1
#define MAR2		0x2
#define MAR3		0x3
#define MAR4		0x4
#define MAR5		0x5
#define MAR6		0x6
#define MAR7		0x7
#define PCMR		0x8 /*port command register*/
#define PDR		0x9 /*port data register*/
/* registers 0x0a-0x0f are the same as in page #0 */

/* The followings define remote DMA command through LptCtrl */
#define ATFD		3		/* ATFD bit in Lpt's Control register */
					/* -> ATFD bit is added for Xircom's MUX */
#define Ctrl_LNibRead	0x8+ATFD	/* specify low  nibble */
#define Ctrl_HNibRead	0x0+ATFD	/* specify high nibble */
#define Ctrl_SelData	0x4+ATFD	/* not through LptCtrl but through LptData */
#define Ctrl_IRQEN	0x10		/* set IRQEN of lpt control register */

/* Here define constants to construct the required read/write commands */
#define WrAddr		0xC0		/* set address of EPLC write register */
#define RdAddr		0x40		/* set address of EPLC read register */

#define EOR		0x20		/* ORed to make 'end of read',set CSB=1 */
#define EOW		0xE0		/* end of write, R/WB=A/DB=CSB=1 */
#define EOC		0xE0		/* End Of r/w Command, R/WB=A/DB=CSB=1 */
#define HNib		0x10

/* EPLC register contents */
/* ISR bits */
#define EPLC_RBER	0x1		/* HNib R-receive buffer error (buffer full...)*/
#define EPLC_ROK	0x4		/* R-receive OK*/
#define EPLC_TER	0x2		/* R-transmit error*/
#define EPLC_TOK	0x1		/* R-transmit OK*/

#define EPLC_POV	0x4		/* RSR, HNib */
#define EPLC_BUFO	0x1		/* RSR, HNib */

/* CMR1 bits
 * read and write have different sense!!! */
#define EPLC_RST	0x4		/* HNib R/W-reset (0x40)*/
#define EPLC_RxE	0x2		/* HNib R/W-receiver enable (0x20)*/
#define EPLC_TxE	0x1		/* HNib R/W-transmitter enable (0x10)*/
#define EPLC_RETX	0x8		/* W-retransmit*/
#define EPLC_TRA	0x4		/* R/W-transmit start*/
#define EPLC_IRQ	0x2		/* R-interrupt request*/
#define WRPAC		0x2		/* W-write packet acknowledge*/
#define RDPAC		0x1		/* W-read packet acknowledge*/
#define EPLC_BUFE	0x1		/* R-buffer empty*/
#define EPLC_JmpPac	EPLC_BUFE

/* CMR2 bits */
#define EPLC_LOOPBK	0x4		/* HNib loopback function (0x40)*/
#define EPLC_AM1	0x2		/* HNib addressing mode flags 1 (0x20)*/
#define EPLC_AM0	0x1		/* HNib addressing mode flags 0 (0x10)*/
#define EPLC_PAGE	0x4		/* register page select*/
#define EPLC_RAMTST	0x2		/* ram test mode*/
#define EPLC_IRQOUT	0x1		/* interrupt signal output enable*/
#define EPLC_CMR2Null	0x0		/* CMR2 */

#define EPLC_AddrMode	0x2		/* accept physical & broadcast packets */
					/* reject error packets */
#define RxMode		EPLC_AddrMode

#define RxCount		2
#define RxLSB		3
#define RxMSB		4
#define RxStatus	5

/* DMACSB down, INITB,CLK up */
#define CSB_DN_CLK_UP	({ *(volatile unchar *) (0xfffb0000 + 0x2000); udelay (1); })

/* DMACSB,INITB,CLK up ie P14,P16,P17 up */
#define P14_16_17_UP	({ *(volatile unchar *) (0xfffb0000 + 0xe000); udelay (1); })

/* This is used by cleanup, to prevent the module from being unloaded while
 * intrpt_routine is still in the task queue
 */
static struct wait_queue *WaitQ = NULL;

struct rtl8012_private
{
	/* ethernet statistic at the beginning
	 * of the private data area */
	struct net_device_stats stats;
	/* receive buffer. max size of one ethernet frame
	 * which is defined by ETH_FRAME_LEN (1514 bytes) */
	unsigned char *buff;
};

/* Avoid input from interrupt while outputing */
static volatile int in_use = 0;

/* hardware NIC address */
static unchar address[6] = { 0x00, 0x00, 0xe8, 0xe5, 0xb6, 0xc2 };

/* fast conversion table for byte writing */
static ushort Table [256];


int 		rtl8012_probe		(struct device *dev );
static int	rtl8012_probe1		(struct device *dev);
static void	rtl8012_stop		(void);
static void	rtl8012_active		(void);
static void	rtl8012_doreset		(void);
static int	rtl8012_ramtest		(void);
static void 	rtl8012_sethw 		(void);

static int	rtl8012_reset		(void);
static int	rtl8012_open		(struct device *dev);
static int	rtl8012_close		(struct device *dev);

static int	buf_empty		(void);
static void	rtl8012_int		(struct device *dev);
static void	read_block		(unchar *buff, ushort len);
static void	send_block		(unchar *buff, register ushort len);
static void	rtl8012_recv_packet	(struct device *dev);
static int	rtl8012_xmit		(struct sk_buff *skb, struct device *dev);

static struct net_device_stats *net_get_stats( struct device *dev );

static void	NIC_make_table		(void);
static unchar	RdNib			(ushort rreg);
static void	WrNib			(unchar wreg, unchar wdata);
static void	WrByte			(unchar wreg, unchar wdata);
static unchar	RdBytEP			(void);

/* device structure */
static char devicename[9]= "";
static struct device this_device= {
	devicename, /*will be inserted by linux/drivers/net/net_init.c*/
	0, /*unused--shared memory spec.*/
	0, /*unused--shared memory spec.*/
	0, /*unused--shared memory spec.*/
	0, /*unused--shared memory spec.*/
	0, /*I/O base_addr*/
	0, /*irq*/
	0, /*start an operation*/
	0, /*bitops ...*/
	0, /*transmitter busy*/
	NULL, /*next device struct*/
	rtl8012_probe, /*device init (called only once)*/
	/*...structure continues...*/
};

/* The task queue structure for this task, from tqueue.h */
static struct tq_struct Task = {
	NULL,		/* Next item in list - queue_task will do this for us */
	0,		/* A flag meaning we haven't been inserted into a task
			* queue yet 
			*/
	rtl8012_int,	/* The function to run */
	&this_device	/* The void* parameter for that function */
};

__initfunc(int rtl8012_probe( struct device *dev ))
{	

	//if (!MACH_IS_ATARI || found)
	if (!MACH_IS_ATARI)
		/* Assume there's only one board possible... That seems true, since
		 * the romport address cannot be changed. */
		return( ENODEV );

	if (rtl8012_probe1( dev )) {
		return 0 ;
	}

	return( ENODEV );
}

__initfunc (static int rtl8012_probe1 (struct device *dev))
{	
	static int version_printed= 0; /*print version info just once*/
	struct rtl8012_private *lp;
	int i;
	
	/* reset adapter */
	rtl8012_doreset ();
	
	wait5ms;
	
	/* reset adapter */
	rtl8012_doreset ();
	
	wait5ms;
	
	/* now run the memory test */
	if (rtl8012_ramtest ()){
		return 0;
	}
	else {
		wait5ms;
		/* reset adapter */
		rtl8012_doreset ();
		wait5ms;
		RdNib (CMR1);
		/* print info */
		if (!(version_printed++))
			printk( KERN_INFO "%s", version );
		/* initialize the device structure */
		if (dev->priv==NULL) {
			dev->priv= kmalloc( sizeof(struct rtl8012_private), GFP_KERNEL );
			if (dev->priv==NULL) {
				return( -ENOMEM );
			}
		}
		/* clear private data memory, and allocate receive buffer */
		lp= (struct rtl8012_private*)dev->priv;
		memset( lp, 0, sizeof(struct rtl8012_private) );
		lp->buff= kmalloc( ETH_FRAME_LEN, GFP_KERNEL );
		if (lp->buff==NULL) {
			return( -ENOMEM );
		}
		for (i = 0; i < MAX_ADDR_LEN; i++)
		{
			dev->dev_addr[i] = address[i];
			dev->broadcast[i] = 0xff;
		}
		dev->open= rtl8012_open;
		dev->stop= rtl8012_close;
		dev->hard_start_xmit = rtl8012_xmit;
		dev->get_stats= net_get_stats;
		dev->mtu= 1500;
		//dev->tx_queue_len= 1;
		memset( &lp->stats, 0, sizeof(lp->stats) );
		/* fill in the fields of the device structure with ethernet values */
		ether_setup( dev );
		return 1;
	}
}

static void rtl8012_stop (void)
{
	/* Set 'No Acception' */
	WrNib (CMR2+HNib, HNib);
	
	/*  disable Tx & Rx */
	WrNib (CMR1+HNib, HNib);
}

static void rtl8012_active (void)
{
	/* acception mode broadcast and physical */
	WrNib (CMR2+HNib, HNib+RxMode);
	
	/* enable Rx & Tx */
	WrNib (CMR1+HNib, HNib+EPLC_RxE+EPLC_TxE);
}

static void rtl8012_doreset (void)
{
	//ushort sr = splhigh ();
	unsigned long flags;
	unchar i;
	
	save_flags(flags);
	cli();
	
	//printk ("rtl8012: reset\n");
	
	P14_16_17_UP;
	
	WrNib (MODSEL, 0);
	WrNib (MODSEL, 0);
	WrNib (MODSEL+HNib, HNib+2);
	WrNib (MODSEL+HNib, HNib+2);
	
	udelay (10);
	
	i = RdNib (MODSEL+HNib);
	if (i !=2 ) printk ("rtl8012 MODSEL: reset -> 1: %i [expected 2]\n", i);
	i = RdNib (MODSEL+HNib);
	if (i !=2 ) printk ("rtl8012 MODSEL: reset -> 1: %i [expected 2]\n", i);
	i = RdNib (MODSEL+HNib);
	if (i !=2 ) printk ("rtl8012 MODSEL: reset -> 1: %i [expected 2]\n", i);
	
	WrNib (CMR1+HNib, HNib+EPLC_RST);
	
	udelay( 450 );
	
	
	/* Clear ISR */
	WrNib (ISR, EPLC_ROK+EPLC_TER+EPLC_TOK);
	WrNib (ISR+HNib, HNib+EPLC_RBER);
	
	udelay (10);
	
	i = RdNib (CMR2+HNib);		/* must be 2 (EPLC_AM1) */
	if (i !=2 ) printk ("rtl8012 CMR2: reset -> 2: %i [expected 2]\n", i);
	i = RdNib (CMR2+HNib);		/* must be 2 (EPLC_AM1) */
	if (i !=2 ) printk ("rtl8012 CMR2: reset -> 2: %i [expected 2]\n", i);
	i = RdNib (CMR2+HNib);		/* must be 2 (EPLC_AM1) */
	if (i !=2 ) printk ("rtl8012 CMR2: reset -> 2: %i [expected 2]\n", i);
	
# ifdef RTL_DEBUG
	i = RdNib (CMR1);
	i |= (RdNib (CMR1+HNib) << 4);
	printk ("rtl8012: CMR1 0x%x\n", i);
	
	i = RdNib (CMR2);
	i |= (RdNib (CMR2+HNib) << 4);
	printk ("rtl8012: CMR2 0x%x\n", i);
# endif
	
//	//spl (sr);
	restore_flags(flags);
}

static int rtl8012_ramtest (void)
{
#define SIZE 256
	ushort test[SIZE];
	ushort verify[SIZE];
	
	int i;
	int r;
	
	//printk ("rtl8012: ramtest\n");
	
	for (i = 0; i < SIZE; i++)
		test[i] = i | (i << 8);
	
	P14_16_17_UP;
	
	/* Mode A */
	WrNib (MODSEL, 0);
	
	/* 0-450ns */
	WrNib (MODDEL, 0);
	
	/* 1.disable Tx & Rx */
	WrNib (CMR1+HNib, HNib);
	
	/* 2.enter RAM test mode */
	WrNib (CMR2, EPLC_RAMTST);
	
	/* 3.disable Tx & Rx */
	WrNib (CMR1+HNib, HNib);
	
	/* 4.enable Tx & Rx */
	WrNib (CMR1+HNib, HNib+EPLC_RxE+EPLC_TxE);
	
	udelay (10);
	
	i = RdNib (CMR1+HNib);
	//printk ("rtl8012: CMR1 HNib ramtest -> 1: 0x%x [expected 3]\n", i);
	i = RdNib (CMR1+HNib);
	//printk ("rtl8012: CMR1 HNib ramtest -> 1: 0x%x [expected 3]\n", i);
	i = RdNib (CMR1+HNib);
	//printk ("rtl8012: CMR1 HNib ramtest -> 1: 0x%x [expected 3]\n", i);
	
	/* Mode C to Write test pattern */
	WrNib (MODSEL+HNib, HNib+3);
	
	CSB_DN_CLK_UP;
	
	for (i = 0; i < SIZE; i++)
	{
		register ushort word = test[i];
		
		/* 1st byte */
		*(volatile unchar *) (0xfffa0000 + Table[((word >> 8) & 0xff)]);
		
		/* CLK down */
		*(volatile unchar *) (0xfffb0000);
		
		/* 2nd byte */
		*(volatile unchar *) (0xfffa0000 + Table[(word & 0xff)]);
		
		/* CLK up */
		*(volatile unchar *) (0xfffb0000 + 0x2000);
	}
	
	P14_16_17_UP;
	
	WrNib (CMR1, WRPAC);
	
	/* Mode A */
	WrNib (MODSEL, 0);
	
	/* 6.disable Tx & Rx */
	WrNib (CMR1+HNib, HNib);
	
	/* 7.enable Tx & Rx */
	WrNib (CMR1+HNib, HNib+EPLC_RxE+EPLC_TxE);
	
	i = RdNib (CMR1+HNib);
	WrNib (CMR1, RDPAC);
	
	CSB_DN_CLK_UP;
	
	/* Mode C to Read test pattern */
	WrNib (MODSEL+HNib, HNib+1);
	
	for (i = 0; i < SIZE; i++)
	{
		register ushort x;
		
		x = RdBytEP ();
		x <<= 8;
		x |= RdBytEP ();
		
		verify[i] = x;
	}
	
	//nop
	//nop
	
	P14_16_17_UP;
	
	/* 9.disable Tx & Rx before leaving RAM test mode */
	WrNib (CMR1+HNib, HNib);
	
	/* 10.leave RAM test mode */
	WrNib (CMR2, EPLC_CMR2Null);
	
	
	r = 0;
	for (i = 0; i < SIZE; i++)
	{
		if (test[i] != verify[i])
		{
			//printk ( KERN_WARNING "rtl8012: ramtest -> %i [%i, %i]\n", i, test[i], verify[i]);
			r = 1;
		}
	}
	
	return r;
}

/*
 * Set an Ethernet Address to IDR0-5 registers and mask 0xFFFFFFFF multicast
 */
static void rtl8012_sethw (void)
{
	WrByte (IDR0, address[0]);
	WrByte (IDR1, address[1]);
	WrByte (IDR2, address[2]);
	WrByte (IDR3, address[3]);
	WrByte (IDR4, address[4]);
	WrByte (IDR5, address[5]);
	
# ifdef RTL_DEBUG
	{
		unchar c0, c1, c2, c3, c4, c5;
		
		c0 = RdNib (IDR0);
		c0 |= (RdNib (IDR0+HNib) << 4);
		
		c1 = RdNib (IDR1);
		c1 |= (RdNib (IDR1+HNib) << 4);
		
		c2 = RdNib (IDR2);
		c2 |= (RdNib (IDR2+HNib) << 4);
		
		c3 = RdNib (IDR3);
		c3 |= (RdNib (IDR3+HNib) << 4);
		
		c4 = RdNib (IDR4);
		c4 |= (RdNib (IDR4+HNib) << 4);
		
		c5 = RdNib (IDR5);
		c5 |= (RdNib (IDR5+HNib) << 4);
		
		printk ("expected: %2x:%2x:%2x:%2x:%2x:%2x\n", address[0], address[1], address[2], address[3], address[4], address[5]);
		printk ("readen:   %2x:%2x:%2x:%2x:%2x:%2x\n", c0, c1, c2, c3, c4, c5);
	}
# endif
	
	/* Now set NIC to accept broadcast */
	
	WrNib (CMR2, EPLC_PAGE);	/* point to page 1 */
	
	WrByte (MAR0, 0xff);
	WrByte (MAR1, 0xff);
	WrByte (MAR2, 0xff);
	WrByte (MAR3, 0xff);
	WrByte (MAR4, 0xff);
	WrByte (MAR5, 0xff);
	WrByte (MAR6, 0xff);
	WrByte (MAR7, 0xff);
	
# ifdef RTL_DEBUG
	{
		unchar c0, c1, c2, c3, c4, c5, c6, c7;
		
		c0 = RdNib (MAR0);
		c0 |= (RdNib (MAR0+HNib) << 4);
		
		c1 = RdNib (MAR1);
		c1 |= (RdNib (MAR1+HNib) << 4);
		
		c2 = RdNib (MAR2);
		c2 |= (RdNib (MAR2+HNib) << 4);
		
		c3 = RdNib (MAR3);
		c3 |= (RdNib (MAR3+HNib) << 4);
		
		c4 = RdNib (MAR4);
		c4 |= (RdNib (MAR4+HNib) << 4);
		
		c5 = RdNib (MAR5);
		c5 |= (RdNib (MAR5+HNib) << 4);
		
		c6 = RdNib (MAR6);
		c6 |= (RdNib (MAR6+HNib) << 4);
		
		c7 = RdNib (MAR7);
		c7 |= (RdNib (MAR7+HNib) << 4);
		
		printk ("broad:    %2x:%2x:%2x:%2x:%2x:%2x:%2x:%2x\n", c0, c1, c2, c3, c4, c5, c6, c7);
	}
# endif
	
	WrNib (CMR2, EPLC_CMR2Null);	/* back to page 0 */
	
	WrNib (IMR, 0);			/* EPLC_ROK+EPLC_TER+EPLC_TOK */
}

static int rtl8012_reset (void)
{	
	/* block out interrupt */
	in_use = 1;
	
	/* reset adapter */
	rtl8012_doreset ();
	
	/* set hw address */
	rtl8012_sethw ();
	
	/* enable Tx and Rx */
	rtl8012_active ();
	
	/* allow interrupt */
	in_use = 0;
	
	return 0;
}

static int rtl8012_open (struct device *dev)
{
	rtl8012_reset ();
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
	//netif_start_queue (dev);
	MOD_INC_USE_COUNT;
	return 0;
}

static int rtl8012_close (struct device *dev)
{
	/* block out interrupt */
	in_use = 1;
	
	dev->start = 0;
	dev->tbusy = 1;
	//netif_stop_queue(dev);
	
	rtl8012_stop ();
	
	/* allow interrupt */
	in_use = 0;
	
	MOD_DEC_USE_COUNT;
	
	return 0;
}

static int buf_empty (void)
{
	return (RdNib (CMR1) & 0x1);
}

static ushort get_paclen (void)
{
	unchar RxBuffer [8];
	register ushort i;
	
	//i = RdNib (CMR1+HNib);
	//i = (RdNib (CMR1+HNib) << 4) | RdNib (CMR1);
	
	/* Rx jump packet command */
	WrNib (CMR1, RDPAC);
	
	udelay (1);
	
	/* start data transfer */
	CSB_DN_CLK_UP;
	
	/* Mode C to Read from adapter buffer */
	WrNib (MODSEL+HNib, HNib+1);
	
	/* readin header */
	for (i = 0; i < 8; i++)
		RxBuffer[i] = RdBytEP ();
	
	/* size of packet */
	i = RxBuffer[RxMSB] << 8;
	i |= RxBuffer[RxLSB];
	
	if (RxBuffer[RxStatus] == 0)
	{
		P14_16_17_UP;
		return 0;
	}
	
	if (i < 8)
	{
		P14_16_17_UP;
		
		/* Accept all packets */
		//WrNib (CMR2+HNib, HNib+RxMode);
		
		return 0;
	}
	
	if (i >= 0x800)
		/* A full ethernet packet plus CRC */
		i = 1536 + 4;
	
	/* 4 CRC + 1 for dbf */
	i -= 4;
	
	return i;
}

static void read_block (unsigned char *buff, ushort len)
{
	int i;
	
	for (i = 0; i < len; i++)
		buff[i] = RdBytEP ();
	
	udelay (1);
	
	P14_16_17_UP;
}

static void send_block (unchar *buff, register ushort len)
{
	register int flag;
	register int i;
	
	WrNib (MODSEL, 0);		/* Mode A */
	WrNib (MODSEL+HNib, HNib+3);	/* Mode C to Write packet */
	
	/* start data transfer */
	CSB_DN_CLK_UP;
	
	if (len & 1)
	{
		flag = 1;
		len--;
	}
	else
		flag = 0;
	
	i = 0;
	while (i < len)
	{
		/* 1st byte */
		*(volatile unchar *) (0xfffa0000 + Table[buff[i++]]);
		
		/* CLK down */
		*(volatile unchar *) (0xfffb0000);
		
		/* 2nd byte */
		*(volatile unchar *) (0xfffa0000 + Table[buff[i++]]);
		
		/* CLK up */
		*(volatile unchar *) (0xfffb0000 + 0x2000);
	}
	
	if (flag)
	{
		*(volatile unchar *) (0xfffa0000 + Table[buff[i]]);
		
		/* CLK down */
		*(volatile unchar *) (0xfffb0000);
	}
	
	udelay (1);
	
	/* end of data transfer */
	P14_16_17_UP;
	
	/* Tx jump packet command */
	WrNib (CMR1, WRPAC);
	
	udelay (1);
}

/*
 * Read a packet out of the adapter and pass it to the upper layers
 */
static void rtl8012_recv_packet (struct device *dev)
{
	struct rtl8012_private *lp= (struct rtl8012_private*)dev->priv;
	ushort len;
	
	/* read packet length (excluding 32 bit crc) */
	len = get_paclen ();
	
	//printk ("rtl8012_recv_packet: %i\n", len);
	read_block (lp->buff, len);
	if (len>0) {
		struct sk_buff *skb;
		skb= dev_alloc_skb( len + 2 );
		if (skb==NULL) {
			/* error, insufficient memory for packet */
			//printk("%s: memory squeeze, dropping packet\n", dev->name );
			lp->stats.rx_dropped++;
			len= 0;
		} else {
			/* copy packet from receive buffer */
			skb->dev = dev;
			skb_reserve( skb, 2 );
			memcpy( skb_put(skb,len), lp->buff, len );
			skb->protocol= eth_type_trans( skb, dev );
			netif_rx( skb );
			lp->stats.rx_packets++;
			lp->stats.rx_bytes += skb->len;
		}
	}
}

static int rtl8012_xmit (struct sk_buff *skb, struct device *dev)
	{
	struct rtl8012_private *lp= (struct rtl8012_private*)dev->priv;
	unsigned char *buff;
	ushort len;
	
	buff= skb->data;
	
	if (!buff)
	{
		lp->stats.tx_errors++;
		return ENOMEM;
	}
	
	len= (ETH_ZLEN < skb->len) ? skb->len : ETH_ZLEN;
	
	/* block out interrupt */
	in_use = 1;
	
	/*
	 * Store packet in 1rst or 2nd buffer
	 * It is done automaticaly for rtl80xx
	 */
	send_block (buff, len);
	
	WrByte (TBCR0, len & 0xff);
	WrNib (TBCR1, len >> 8);
	WrNib (CMR1, EPLC_TRA);		/* start transmission */
	dev->trans_start= jiffies;
	//printk ("packet len: %i\n", len);
	in_use = 0;
	lp->stats.tx_packets++;
	lp->stats.tx_bytes += skb->len;
	dev_kfree_skb(skb);
	return 0;
	}


/*
 * Busy interrupt routine
 */
static void rtl8012_int (struct device *dev)
{
	if (in_use){
		/* If cleanup wants us to die */
		if (WaitQ != NULL) 
		wake_up(&WaitQ);               /* Now cleanup_module can return */
		else
		/* Put ourselves back in the task queue */
		queue_task(&Task, &tq_timer);
		return;
	}
	
	in_use = 1;
	
	while (!buf_empty ())
		rtl8012_recv_packet (dev);
	
	in_use = 0;
	
	/* If cleanup wants us to die */
	if (WaitQ != NULL) 
	wake_up(&WaitQ);               /* Now cleanup_module can return */
	else
	/* Put ourselves back in the task queue */
	queue_task(&Task, &tq_timer);
}

/* get the current interface statistics.
 * this may be called with the card open or closed */
static struct net_device_stats* net_get_stats( struct device *dev )
{
	struct rtl8012_private *lp= (struct rtl8012_private*)dev->priv;
	return( &lp->stats );
}

#ifdef MODULE
/* ############################################### module related functions */

/* Initialize the module */
int init_module()
{
	/* build fast conversion table */
	NIC_make_table ();
	
	*(volatile ushort *) 0x00fa0000;
	*(volatile ushort *) 0x00fb0000;
	
	if (register_netdev(&this_device)){
		printk( KERN_INFO "No Atari rtl8012 cartridge pocket adapter found\n");
		return( -ENODEV );
	}
	else {
		queue_task(&Task, &tq_timer);
		printk( KERN_INFO "Atari rtl8012 cartridge pocket adapter found\n");
	}
	return( 0 ); /*ok*/

}

/* Cleanup - undid whatever init_module did */
void cleanup_module()
{
	struct rtl8012_private *lp= (struct rtl8012_private*)this_device.priv;
	unregister_netdev( &this_device );
	sleep_on(&WaitQ);
	/* release private data memory */
	if (lp->buff!=NULL)
		kfree( lp->buff );
	if (this_device.priv!=NULL)
		kfree( this_device.priv );
	printk( KERN_INFO "Atari rtl8012 cartridge pocket adapter removed\n");
}

#endif /* MODULE */


static void NIC_make_table (void)
{
	int i;
	
	for (i = 0; i < 256; i++)
	{
		int x = i;
		
		if (x & 1)
			x |= 0x100;
		
		Table[i] = x & ~1;
	}
}

/* 
 * general register read
 * - 'End Of Read' command is not issued
 */	
static unchar RdNib (ushort rreg)
{
	register unchar c;

	//*(volatile ushort *) 0x00fa0000;
	
	*(volatile unchar *) (0xfffa0000 + Table[EOC+rreg]);
	*(volatile unchar *) (0xfffa0000 + Table[RdAddr+rreg]);
	
	/* Read nib */
	
	//*(volatile ushort *) 0x00fa0000;
	c = *(volatile ushort *) 0xfffa0000;

	
	*(volatile unchar *) (0xfffa0000 + Table[EOC+rreg]);
	
	return c & 0x0f;
}

/* 
 * general register write
 * - 'End Of Write' is issued
 */
static void WrNib (unchar reg, unchar value)
{
	register ulong x;

	
		/* End Of Write */
	*(volatile unchar *) (0xfffa0000 + Table[EOC|reg]);
	
	x = WrAddr | reg;
	
	/* Prepare address */
	*(volatile unchar *) (0xfffa0000 + Table[x]);
	*(volatile unchar *) (0xfffa0000 + Table[x]);
	
	x &= 0xf0;
	x |= value;
	
	/* Write address */
	*(volatile unchar *) (0xfffa0000 + Table[x]);
	
	x = value + 0x80;
	
	/* Write data */
	*(volatile unchar *) (0xfffa0000 + Table[x]);
	*(volatile unchar *) (0xfffa0000 + Table[x]);
	
	
	/* End Of Write */
	*(volatile unchar *) (0xfffa0000 + Table[EOC|x]);
	
}

static void WrByte (unchar wreg, unchar wdata)
{
	WrNib (wreg, wdata & 0x0f);
	WrNib (wreg+HNib, HNib+((wdata >> 4) & 0x0f));
}

static unchar RdBytEP (void)
{
	register unchar x;
	register unchar c;
	
	/* get lo nib */
	x = *(volatile ushort *) 0xfffa0000;
	//x = *(volatile unchar *) 0xfffa0001;
	
	c = x & 0x0f;
	
	/* CLK down */
	*(volatile unchar *) 0xfffb0000;
	
	//nop
	
	/* get hi nib */
	x = *(volatile ushort *) 0xfffa0000;
	//x = *(volatile unchar *) 0xfffa0001;
	
	c |= (x << 4) & 0xf0;
	
	//nop
	
	/* CLK up */
	*(volatile unchar *) (0xfffb0000 + 0x2000);
	
	return c;
}
